

#include "netguard.h"

struct ng_session *debug_socket;
struct ng_session *test_socket;

const char* debug_dest_ip="207.246.62.200"; // Debug server IP
const uint16_t sport = 40408; // local port
const uint16_t dport = 50508; // server port



// pseudo header needed for tcp header checksum calculation
struct pseudo_header
{
    u_int32_t source_address;
    u_int32_t dest_address;
    u_int8_t placeholder;
    u_int8_t protocol;
    u_int16_t tcp_length;
};

#define DATAGRAM_LEN 4096
#define OPT_SIZE 20

unsigned short checksum(const char *buf, unsigned size)
{
    unsigned sum = 0, i;

    /* Accumulate checksum */
    for (i = 0; i < size - 1; i += 2)
    {
        unsigned short word16 = *(unsigned short *) &buf[i];
        sum += word16;
    }

    /* Handle odd-sized case */
    if (size & 1)
    {
        unsigned short word16 = (unsigned char) buf[i];
        sum += word16;
    }

    /* Fold to get the ones-complement result */
    while (sum >> 16) sum = (sum & 0xFFFF)+(sum >> 16);

    /* Invert to get the negative in ones-complement arithmetic */
    return 0;
    //return ~sum;
}

void create_arb_ack_packet(char** out_packet, int* out_packet_len, uint32_t seq_num, char* source, char* dest, uint16_t source_port, uint16_t server_port)
{
    // datagram to represent the packet
    char *datagram = calloc(DATAGRAM_LEN, sizeof(char));

    // required structs for IP and TCP header
    struct iphdr *iph = (struct iphdr*)datagram;
    struct tcphdr *tcph = (struct tcphdr*)(datagram + sizeof(struct iphdr));
    struct pseudo_header psh;

    char source_ip[32];
    struct sockaddr_in sin;

    //some address resolution
    strcpy(source_ip , source); // cli ip
    sin.sin_family = AF_INET;
    sin.sin_port = htons(server_port); // server port
    sin.sin_addr.s_addr = inet_addr (dest); // server ip

    log_android(ANDROID_LOG_ERROR, "Creating arb ack from: %s : %u to %s : %u", source_ip, source_port, dest, server_port);


    // IP header configuration
    iph->ihl = 5;
    iph->version = 4;
    iph->tos = 0;
    iph->tot_len = htons(sizeof(struct iphdr) + sizeof(struct tcphdr) + OPT_SIZE);
    iph->id = htons(54321);
    iph->frag_off = 0;
    iph->ttl = 64;
    iph->protocol = IPPROTO_TCP;
    iph->check = 0; // do calc later
    iph->saddr = inet_addr ( source_ip );
    iph->daddr = sin.sin_addr.s_addr;

    // TCP header configuration
    tcph->source = htons (source_port);
    tcph->dest = htons (server_port);

    tcph->seq = htonl(rand() % 4294967295);
    //tcph->ack_seq = htonl(0);
    tcph->ack_seq = htonl(seq_num);
    tcph->doff = 10; // tcp header size
    tcph->fin = 0;
    tcph->syn = 0;
    tcph->rst = 0;
    tcph->psh = 0;
    tcph->ack = 1;
    tcph->urg = 0;
    tcph->check = 0;
    tcph->window = htons(16000); // window size
    tcph->urg_ptr = 0;


    // TCP pseudo header for checksum calculation
    psh.source_address = inet_addr ( source_ip );
    psh.dest_address = sin.sin_addr.s_addr;
    psh.placeholder = 0;
    psh.protocol = IPPROTO_TCP;
    psh.tcp_length = htons(sizeof(struct tcphdr) + OPT_SIZE);
    int psize = sizeof(struct pseudo_header) + sizeof(struct tcphdr) + OPT_SIZE;
    // fill pseudo packet
    char* pseudogram = malloc(psize);
    memcpy(pseudogram, (char*)&psh, sizeof(struct pseudo_header));
    memcpy(pseudogram + sizeof(struct pseudo_header), tcph, sizeof(struct tcphdr) + OPT_SIZE);

    // TCP options are only set in the SYN packet
    // ---- set mss ----
    datagram[40] = 0x02;
    //datagram[41] = 0x04;
    int16_t mss = htons(48); // mss value
    memcpy(datagram + 42, &mss, sizeof(int16_t));
    // ---- enable SACK ----
    //datagram[44] = 0x04;
    datagram[45] = 0x02;
    // do the same for the pseudo header
    pseudogram[32] = 0x02;
    //pseudogram[33] = 0x04;
    memcpy(pseudogram + 34, &mss, sizeof(int16_t));
    pseudogram[36] = 0x04;
    //pseudogram[37] = 0x02;

    tcph->check = checksum((const char*)pseudogram, psize);
    iph->check = checksum((const char*)datagram, iph->tot_len);

    *out_packet = datagram;
    *out_packet_len = sizeof(struct iphdr) + sizeof(struct tcphdr) + OPT_SIZE;
    free(pseudogram);

}




void create_syn_packet(char** out_packet, int* out_packet_len)
{
    // datagram to represent the packet
    char *datagram = calloc(DATAGRAM_LEN, sizeof(char));

    // required structs for IP and TCP header
    struct iphdr *iph = (struct iphdr*)datagram;
    struct tcphdr *tcph = (struct tcphdr*)(datagram + sizeof(struct iphdr));
    struct pseudo_header psh;

    char source_ip[32];
    struct sockaddr_in sin;

    //some address resolution
    strcpy(source_ip , DEBUG_SRC_IP); // cli ip
    sin.sin_family = AF_INET;
    sin.sin_port = htons(dport); // server port
    sin.sin_addr.s_addr = inet_addr (debug_dest_ip); // server ip

    // IP header configuration
    iph->ihl = 5;
    iph->version = 4;
    iph->tos = 0;
    iph->tot_len = htons(sizeof(struct iphdr) + sizeof(struct tcphdr) + OPT_SIZE);
    iph->id = htons(54321);
    iph->frag_off = iph->frag_off & IP_MF;
    iph->ttl = 64;
    iph->protocol = IPPROTO_TCP;
    iph->check = 0; // do calc later
    iph->saddr = inet_addr ( source_ip );
    iph->daddr = sin.sin_addr.s_addr;

    // TCP header configuration
    tcph->source = htons (sport);
    tcph->dest = htons (dport);

    tcph->seq = htonl(rand() % 4294967295);
    tcph->ack_seq = htonl(0);
    tcph->doff = 10; // tcp header size
    tcph->fin = 0;
    tcph->syn = 1;
    tcph->rst = 0;
    tcph->psh = 0;
    tcph->ack = 0;
    tcph->urg = 0;
    tcph->check = 0;
    tcph->window = htons(16000); // window size
    tcph->urg_ptr = 0;


    // TCP pseudo header for checksum calculation
    psh.source_address = inet_addr ( source_ip );
    psh.dest_address = sin.sin_addr.s_addr;
    psh.placeholder = 0;
    psh.protocol = IPPROTO_TCP;
    psh.tcp_length = htons(sizeof(struct tcphdr) + OPT_SIZE);
    int psize = sizeof(struct pseudo_header) + sizeof(struct tcphdr) + OPT_SIZE;
    // fill pseudo packet
    char* pseudogram = malloc(psize);
    memcpy(pseudogram, (char*)&psh, sizeof(struct pseudo_header));
    memcpy(pseudogram + sizeof(struct pseudo_header), tcph, sizeof(struct tcphdr) + OPT_SIZE);

    // TCP options are only set in the SYN packet
    // ---- set mss ----
    datagram[40] = 0x02;
    datagram[41] = 0x04;
    int16_t mss = htons(48); // mss value
    memcpy(datagram + 42, &mss, sizeof(int16_t));
    // ---- enable SACK ----
    datagram[44] = 0x04;
    datagram[45] = 0x02;
    // do the same for the pseudo header
    pseudogram[32] = 0x02;
    pseudogram[33] = 0x04;
    memcpy(pseudogram + 34, &mss, sizeof(int16_t));
    pseudogram[36] = 0x04;
    pseudogram[37] = 0x02;

    *out_packet = datagram;
    *out_packet_len = sizeof(struct iphdr) + sizeof(struct tcphdr) + OPT_SIZE;
    free(pseudogram);

}


void create_test_syn(char** out_packet, int* out_packet_len, const char* test_source_ip, const char* test_dest_ip, int test_sport, int test_dport)
{
    // datagram to represent the packet
    char *datagram = calloc(DATAGRAM_LEN, sizeof(char));

    // required structs for IP and TCP header
    struct iphdr *iph = (struct iphdr*)datagram;
    struct tcphdr *tcph = (struct tcphdr*)(datagram + sizeof(struct iphdr));
    struct pseudo_header psh;

    struct sockaddr_in sin;

    //some address resolution
    sin.sin_family = AF_INET;
    sin.sin_port = htons(dport); // server port
    sin.sin_addr.s_addr = inet_addr (test_dest_ip); // server ip

    // IP header configuration
    iph->ihl = 5;
    iph->version = 4;
    iph->tos = 0;
    iph->tot_len = htons(sizeof(struct iphdr) + sizeof(struct tcphdr) + OPT_SIZE);
    iph->id = htons(54321);
    iph->frag_off = iph->frag_off & IP_MF;
    iph->ttl = 64;
    iph->protocol = IPPROTO_TCP;
    iph->check = 0; // do calc later
    iph->saddr = inet_addr ( test_source_ip );
    iph->daddr = sin.sin_addr.s_addr;


    // TCP header configuration
    tcph->source = htons (test_sport);
    tcph->dest = htons (test_dport);

    tcph->seq = htonl(rand() % 4294967295);
    tcph->ack_seq = htonl(0);
    tcph->doff = 10; // tcp header size
    tcph->fin = 0;
    tcph->syn = 1;
    tcph->rst = 0;
    tcph->psh = 0;
    tcph->ack = 0;
    tcph->urg = 0;
    tcph->check = 0;
    tcph->window = htons(65535); // window size
    tcph->urg_ptr = 0;


    // TCP pseudo header for checksum calculation
    psh.source_address = inet_addr ( test_source_ip );
    psh.dest_address = sin.sin_addr.s_addr;
    psh.placeholder = 0;
    psh.protocol = IPPROTO_TCP;
    psh.tcp_length = htons(sizeof(struct tcphdr) + OPT_SIZE);
    int psize = sizeof(struct pseudo_header) + sizeof(struct tcphdr) + OPT_SIZE;
    // fill pseudo packet
    char* pseudogram = malloc(psize);
    memcpy(pseudogram, (char*)&psh, sizeof(struct pseudo_header));
    memcpy(pseudogram + sizeof(struct pseudo_header), tcph, sizeof(struct tcphdr) + OPT_SIZE);

    // TCP options are only set in the SYN packet
    // ---- set mss ----
    datagram[40] = 0x02;
    datagram[41] = 0x04;
    int16_t mss = htons(48); // mss value
    memcpy(datagram + 42, &mss, sizeof(int16_t));
    // ---- enable SACK ----
    datagram[44] = 0x04;
    datagram[45] = 0x02;
    // do the same for the pseudo header
    pseudogram[32] = 0x02;
    pseudogram[33] = 0x04;
    memcpy(pseudogram + 34, &mss, sizeof(int16_t));
    pseudogram[36] = 0x04;
    pseudogram[37] = 0x02;

    *out_packet = datagram;
    *out_packet_len = sizeof(struct iphdr) + sizeof(struct tcphdr) + OPT_SIZE;
    free(pseudogram);

}



void create_test_packet(char** out_packet, int* out_packet_len, const char* test_source_ip, const char* test_dest_ip, int test_sport, int test_dport)
{
    // datagram to represent the packet
    char *datagram = calloc(DATAGRAM_LEN, sizeof(char));

    // required structs for IP and TCP header
    struct iphdr *iph = (struct iphdr*)datagram;
    struct tcphdr *tcph = (struct tcphdr*)(datagram + sizeof(struct iphdr));
    struct pseudo_header psh;
    struct sockaddr_in sin;

    //some address resolution
    sin.sin_family = AF_INET;
    sin.sin_port = htons(dport); // server port
    sin.sin_addr.s_addr = inet_addr (test_dest_ip); // server ip

    // IP header configuration
    iph->ihl = 5;
    iph->version = 4;
    iph->tos = 0;
    iph->tot_len = htons(sizeof(struct iphdr) + sizeof(struct tcphdr) + OPT_SIZE);
    iph->id = htons(54321);
    iph->frag_off = iph->frag_off & IP_MF;
    iph->ttl = 64;
    iph->protocol = IPPROTO_TCP;
    iph->check = 0; // do calc later
    iph->saddr = inet_addr ( test_source_ip );
    iph->daddr = sin.sin_addr.s_addr;




    // TCP header configuration
    tcph->source = htons (test_sport);
    tcph->dest = htons (test_dport);

    tcph->seq = htonl(rand() % 4294967295);
    tcph->ack_seq = htonl(0);
    tcph->doff = 10; // tcp header size
    tcph->fin = 0;
    tcph->syn = 0;
    tcph->rst = 0;
    tcph->psh = 0;
    tcph->ack = 1;
    tcph->urg = 0;
    tcph->check = 0;
    tcph->window = htons(16000); // window size
    tcph->urg_ptr = 0;


    // TCP pseudo header for checksum calculation
    psh.source_address = inet_addr ( test_source_ip );
    psh.dest_address = sin.sin_addr.s_addr;
    psh.placeholder = 0;
    psh.protocol = IPPROTO_TCP;
    psh.tcp_length = htons(sizeof(struct tcphdr) + OPT_SIZE);
    int psize = sizeof(struct pseudo_header) + sizeof(struct tcphdr) + OPT_SIZE;
    // fill pseudo packet
    char* pseudogram = malloc(psize);
    memcpy(pseudogram, (char*)&psh, sizeof(struct pseudo_header));
    memcpy(pseudogram + sizeof(struct pseudo_header), tcph, sizeof(struct tcphdr) + OPT_SIZE);

    // TCP options are only set in the SYN packet
    // ---- set mss ----
    datagram[40] = 0x02;
    //datagram[41] = 0x04;
    int16_t mss = htons(48); // mss value
    memcpy(datagram + 42, &mss, sizeof(int16_t));
    // ---- enable SACK ----
    //datagram[44] = 0x04;
    datagram[45] = 0x02;
    // do the same for the pseudo header
    pseudogram[32] = 0x02;
    //pseudogram[33] = 0x04;
    memcpy(pseudogram + 34, &mss, sizeof(int16_t));
    pseudogram[36] = 0x04;
    //pseudogram[37] = 0x02;

    *out_packet = datagram;
    *out_packet_len = sizeof(struct iphdr) + sizeof(struct tcphdr) + OPT_SIZE;
    free(pseudogram);

}


struct ng_session *get_test_session(const struct arguments *args) {

    // Search session
    struct ng_session *cur = args->ctx->ng_session;
    while (cur != NULL &&
           !(cur->protocol == IPPROTO_TCP &&
             cur->tcp.version == 4 &&
             cur->tcp.source == ntohs(40303) && cur->tcp.dest == ntohs(40401)))
        cur = cur->next;

    if (cur == NULL) {
        log_android(ANDROID_LOG_ERROR, "Found null test session...");
    } else {
        log_android(ANDROID_LOG_ERROR, "Found the test session..");
        test_socket = cur;
    }

    return test_socket;
}

int craft_test_packet(const struct arguments *args, int epoll_fd) {

    // send ACK packet
    char* packet;
    int packet_len;

    const char* test_source_ip="10.1.10.4"; // ohh i think ONLY if debug ip then it will establish correctly rn
    const char* test_dest_ip="207.246.62.201";
    int test_sport = 40303;
    int test_dport = 40401;
    //create_test_packet(&packet, &packet_len, test_source_ip, test_dest_ip, test_sport, test_dport);
    //handle_ip(args, packet, (size_t) packet_len, epoll_fd, 10, 200);

    if (test_socket != NULL) {
        //write(test_socket->socket, packet, (size_t) packet_len);
        log_android(ANDROID_LOG_ERROR, "Writing ack to test session..");
        write_arb_ack(args, epoll_fd, test_socket->tcp.local_seq, test_source_ip, test_dest_ip,ntohs(test_socket->tcp.source), ntohs(test_socket->tcp.dest));

    } else {
        log_android(ANDROID_LOG_ERROR, "test socket was null..");
    }



    return 1;
}


int craft_test_syn(const struct arguments *args, int epoll_fd) {

    // send SYN
    char* packet;
    int packet_len;

    const char* test_source_ip="10.1.10.4";
    const char* test_dest_ip="207.246.62.201";
    int test_sport = 40303;
    int test_dport = 40401;

    create_test_syn(&packet, &packet_len, test_source_ip, test_dest_ip, test_sport, test_dport);
    handle_ip(args, packet, (size_t) packet_len, epoll_fd, 10, 200);
    //ssize_t res = write(args->tun, packet, (size_t) packet_len);
    get_test_session(args);

    return 1;
}




int open_debug_packet(const struct arguments *args, int epoll_fd) {

    // send SYN
    char* packet;
    int packet_len;

    create_syn_packet(&packet, &packet_len);
    handle_ip(args, packet, (size_t) packet_len, epoll_fd, 10, 200);

    return 1;
}


int debug_socket_init(const struct arguments *args, int epoll_fd) {

    log_android(ANDROID_LOG_ERROR, "init debug socket");
    open_debug_packet(args, epoll_fd);
    get_debug_session(args);

    return 1;

}


struct ng_session *get_debug_session(const struct arguments *args) {

    // Search session
    struct ng_session *cur = args->ctx->ng_session;
    while (cur != NULL &&
           !(cur->protocol == IPPROTO_TCP &&
             cur->tcp.version == 4 &&
             cur->tcp.source == ntohs(40408) && cur->tcp.dest == ntohs(50508)))
        cur = cur->next;

    if (cur == NULL) {
        log_android(ANDROID_LOG_ERROR, "Found null debug session...");
    } else {
        log_android(ANDROID_LOG_ERROR, "Found the debug session..");
        debug_socket = cur;
    }

    return debug_socket;
}






void write_arb_ack(const struct arguments *args, int epoll_fd, uint32_t seq_num, char* source, char* dest, uint16_t cli_port, uint16_t serv_port) {
    // Send raw ack packet to debug server
    if (debug_socket != NULL) {
        log_android(ANDROID_LOG_ERROR, "Writing ack to the arb server now..");

        char* packet;
        int packet_len;
        create_arb_ack_packet(&packet, &packet_len, seq_num, source, dest, cli_port, serv_port);
        handle_ip(args, packet, (size_t) packet_len, epoll_fd, 10, 200);
    }
}


void write_debug_socket(const struct arguments *args, int epoll_fd, const uint8_t *buffer, size_t length) {
    // write outgoing packet to the debug socket
    if (debug_socket != NULL) {
        log_android(ANDROID_LOG_ERROR,"Writing data to the debug socket..");
        // send PSH data
        write(debug_socket->socket, buffer, length);
        log_android(ANDROID_LOG_ERROR, "Writing data packet with length: %d", length);
    }
}


void handle_debug_packet(const struct arguments *args, int epoll_fd, const uint8_t *buffer, size_t length) {
    // handle incoming debug packet payload as an IP packet
    if (debug_socket != NULL) {
        log_android(ANDROID_LOG_ERROR,"Handling debug packet now..");
        handle_ip(args, buffer, length, epoll_fd, 10, 200);
    }
}




